/******************************************************************
 *  Character Device Driver: charDev4.c                           *
 *  gcc -O2 -DMODULE -D__KERNEL__ -W -Wall -Wstrict-prototypes    *
 *  -Wmissing-prototypes -isystem                                 *
 *  /lib/modules/`uname -r`/build/include                         *
 *  -c -o charDev5.o charDev5.c                                   *
 *  ***************************************************************/

#include <linux/kernel.h>  /* Needed for KERN_ALERT */
#include <linux/module.h>  /* Needed by all modules */
#include <linux/fs.h>
#include <asm-i386/current.h>
#include <asm-i386/uaccess.h>
#include <asm-i386/page.h>
#include <asm-i386/pgtable.h>
#include <asm-i386/pgtable-2level.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include "cd4.h"

/*******************************************/
#define CHAR_DEV_NAME "myCharDev"
#define MAXBUFF 4096

/*******************************************/
 static char messageBuff[MAXBUFF] = {0} ;

 static int myOpen(struct inode *, struct file *) ;
 static ssize_t myRead(struct file *, char *, size_t, loff_t *) ;
 static int myRelease(struct inode *, struct file *) ;
 static int majorNo ;  // Major device number
 static struct file_operations fileOps = {  // linux/fs.h
          open:  myOpen,
         read: myRead, 
        // write: myWrite,
         release: myRelease
 } ; // charDev operation names - pointer to functions

int init_module(void) {
    printk("Inside the init_module\n");
    majorNo = register_chrdev(0, CHAR_DEV_NAME, &fileOps) ;
                               // linux/fs.h
    if(majorNo < 0) {
        printk("Registration failed - majorNo: %d\n", majorNo) ;
        return majorNo ;
    }
    printk("Make a node: mknod /dev/myCharDev c %d 0\n", majorNo) ;
    return 0;
}
void cleanup_module(void) {
    unregister_chrdev(majorNo, CHAR_DEV_NAME) ;
     printk("Inside the cleanup_module\n");
    
}  

static int myOpen(struct inode *indP, struct file *fP) {
    printk("<1>Inside myOpen\n") ;
    return 0 ;
}

static ssize_t myRead(    struct file *filp,
            char *buffer,    // Buffer pointer
            size_t length,   // Length of the data string
            loff_t *offset) { // Offset

    int cr3, count = 0, i, size ;
    char *mp, *rp ;
    pgd_t *pgdP ;   // structure defined in asm-i386/page.h
    pageDirBase pdb ;

    printk("<1>Inside myRead\n") ;
    pgdP = current -> mm -> pgd ;    // task_struct and mm_struct
    asm __volatile__ (
                "movl %%cr3, %%eax \n\t"
                : "=a" (cr3)
    ) ;

    pdb.cr3  = (unsigned) cr3 ;
    pdb.pgdP = (unsigned) pgdP ;
    
    mp = messageBuff ;
    rp = (char *) &pdb ;
    size = sizeof(pageDirBase) ;
    for(i=0; i<size; ++i) mp[i] = rp[i] ;

// Kernel space to user space
    count = size ;
    mp = messageBuff ;
    for(i=0; i<count; ++i) {
          put_user(*mp, buffer) ;
          mp++; buffer++ ;
    }
    return 0 ;
}


static int myRelease(struct inode *indP, struct file *fP) {
        printk("Inside myRelease\n") ;
    return 0 ;
}


