/******************************************************************
 *  Character Device Driver: charDev6.c                           *
 *  gcc -O2 -DMODULE -D__KERNEL__ -W -Wall -Wstrict-prototypes    *
 *  -Wmissing-prototypes -isystem                                 *
 *  /lib/modules/`uname -r`/build/include                         *
 *  -c -o charDev6.o charDev6.c                                   *
 *  The user program is user6.c                                   *
 *  ***************************************************************/

#include <linux/kernel.h>  /* Needed for KERN_ALERT */
#include <linux/module.h>  /* Needed by all modules */
#include <linux/fs.h>
#include <asm-i386/current.h>
#include <asm-i386/uaccess.h>
#include <asm-i386/page.h>
#include <asm-i386/pgtable.h>
#include <asm-i386/pgtable-2level.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include "cd5.h"

/*******************************************/
#define CHAR_DEV_NAME "myCharDev"
#define MAXBUFF 4096

/*******************************************/
 static char messageBuff[MAXBUFF] = {0} ;

 static int myOpen(struct inode *, struct file *) ;
 static ssize_t myRead(struct file *, char *, size_t, loff_t *) ;
 static int myRelease(struct inode *, struct file *) ;
 static int majorNo ;  // Major device number
 static struct file_operations fileOps = {  // linux/fs.h
         open:  myOpen,
         read: myRead, 
        // write: myWrite,
         release: myRelease
 } ; // charDev operation names - pointer to functions

int init_module(void) {
    printk("Inside the init_module\n");
    majorNo = register_chrdev(0, CHAR_DEV_NAME, &fileOps) ;
                               // linux/fs.h
    if(majorNo < 0) {
        printk("Registration failed - majorNo: %d\n", majorNo) ;
        return majorNo ;
    }
    printk("Make a node: mknod /dev/myCharDev c %d 0\n", majorNo) ;
    return 0;
}
void cleanup_module(void) {
    unregister_chrdev(majorNo, CHAR_DEV_NAME) ;
    printk("Inside the cleanup_module\n");
    
}  

static int myOpen(struct inode *indP, struct file *fP) {
           printk("<1>Inside myOpen\n") ;
           return 0 ;
}

static ssize_t myRead(    struct file *filp,
            char *buffer,    // Buffer pointer
            size_t length,   // Length of the data string
            loff_t *offset) { // Offset

    int count, kAddr, i, size ;
    unsigned char *mp, *kp ;
    pgd_t *pgdP, *pgdOff ;   // pgd_t and pte_t are defined in
    pte_t *pteOff ;          // asm-i386/page.h
    pagetable t ;
    
    printk("<1>Inside myRead\n") ;
    for(i=0; i<MAXBUFF; ++i) messageBuff[i] = '\0' ;    

    sscanf(buffer, "%u", &t.inAddr) ;

    pgdP = current -> mm -> pgd ;  // task_struct and mm_struct are 
                                   // defined in linux/sched.h
                                   // current is defined in
                                   // asm-i386/current.h
    
    t.pgdP = (unsigned) pgdP ;     // Directory base
    asm __volatile__ (
                "movl %%cr3, %%eax \n\t"
                : "=a" (t.cr3)
    ) ;
    
    pgdOff = (pgd_t *) ((int) pgdP | ((t.inAddr & 0xffc00000) >> 20)) ;
                  // Higher order 10 bits * 4 of inAddr
                  // pgdOff is a pointer of type pgd_t defined in
                  // asm-i386/page.h. It is of type
                  // typedef struct { unsigned long long pgd; } pgd_t;
                  // 
     t.pgdOff = (unsigned) pgdOff  ;        // Directory offset
     t.pgdEnt = (unsigned) pgdOff -> pgd ;  // Directory entry

     pteOff = (pte_t *) (0xc0000000 | ((int) pgdOff -> pgd & 0xfffff000) 
                    | ((t.inAddr & 0x003ff000) >> 10)) ;
                  // Kernel virtual address 0xc.....
                  // Next 10 bits  * 4 of inAddr
                  // pteOff is a pointer of type pte_t defined in
                  // asm-i386/page.h. It is of type 
                  // typedef struct { unsigned long pte_low; } pte_t;
                  // 
   
     t.pteOff = (unsigned) pteOff ; 
    
     if(pte_present(*pteOff)) {
	  t.pEnt = pteOff -> pte_low ;
          kAddr = (t.inAddr & 0x00000fff) | 
            (pteOff -> pte_low & 0xfffff000) | 0xc0000000 ;
          t.kAddr = (unsigned) kAddr ;

          kp = (char *) kAddr ;
          for(i=0; i<SIZE; ++i) t.code[i] = kp[i] ;
     }

     mp = messageBuff ;
     kp = (char *) &t ;
     size = sizeof(pagetable) ;
     for(i=0; i<size; ++i) mp[i] = kp[i] ;

// Kernel space to user space
     count = size ;
     mp = messageBuff ;
     for(i=0; i<count; ++i) {
          put_user(*mp, buffer) ;
          mp++; buffer++ ;
     }
    return 0 ;
}


static int myRelease(struct inode *indP, struct file *fP) {
        printk("Inside myRelease\n") ;
    return 0 ;
}


